# -*- coding: utf-8 -*-

import enum
import six

import attr

from cryptodatahub.common.grade import (
    AttackType,
    Grade,
    GradeableComplex,
    GradeableVulnerabilities,
    Vulnerability,
)
from cryptodatahub.common.types import (
    CryptoDataEnumBase,
    CryptoDataEnumOIDBase,
    CryptoDataParamsEnumString,
    CryptoDataParamsNamed,
    CryptoDataParamsOIDOptional,
    convert_enum,
)


@attr.s(frozen=True)
class AuthenticationParams(CryptoDataParamsOIDOptional, GradeableVulnerabilities):
    anonymous = attr.ib(validator=attr.validators.instance_of(bool))

    @classmethod
    def get_gradeable_name(cls):
        return 'authentication'


@attr.s(frozen=True)
class BlockCipherParams(CryptoDataParamsNamed, GradeableVulnerabilities):
    key_size = attr.ib(validator=attr.validators.instance_of(int))
    block_size = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(int)))

    @classmethod
    def get_gradeable_name(cls):
        return 'block cipher'


@attr.s(frozen=True)
class BlockCipherModeParams(CryptoDataParamsNamed, GradeableVulnerabilities):
    @classmethod
    def get_gradeable_name(cls):
        return 'block cipher mode'


@attr.s(frozen=True)
class HashParams(CryptoDataParamsOIDOptional, GradeableVulnerabilities):
    digest_size = attr.ib(validator=attr.validators.instance_of(int))

    @classmethod
    def get_gradeable_name(cls):
        return 'hash'


@attr.s(frozen=True)
class KeyExchangeParams(CryptoDataParamsNamed, GradeableVulnerabilities):
    forward_secret = attr.ib(validator=attr.validators.instance_of(bool))

    @classmethod
    def get_gradeable_name(cls):
        return 'key exchange'


NamedGroupType = enum.Enum('NamedGroupType', 'ELLIPTIC_CURVE FINITE_FIELD HYBRID_PQS')


@attr.s(frozen=True)
class NamedGroupParams(CryptoDataParamsOIDOptional, GradeableVulnerabilities):
    size = attr.ib(validator=attr.validators.instance_of(int))
    group_type = attr.ib(
        converter=convert_enum(NamedGroupType),
        validator=attr.validators.instance_of(NamedGroupType),
    )

    @classmethod
    def get_gradeable_name(cls):
        return 'named group'


Authentication = CryptoDataEnumOIDBase('Authentication', CryptoDataEnumOIDBase.get_json_records(AuthenticationParams))
BlockCipher = CryptoDataEnumBase('BlockCipher', CryptoDataEnumBase.get_json_records(BlockCipherParams))
BlockCipherMode = CryptoDataEnumBase('BlockCipherMode', CryptoDataEnumBase.get_json_records(BlockCipherModeParams))
Hash = CryptoDataEnumOIDBase('Hash', CryptoDataEnumOIDBase.get_json_records(HashParams))
KeyExchange = CryptoDataEnumBase('KeyExchange', CryptoDataEnumBase.get_json_records(KeyExchangeParams))
NamedGroup = CryptoDataEnumOIDBase('NamedGroup', CryptoDataEnumOIDBase.get_json_records(NamedGroupParams))


@attr.s(frozen=True)
class MACParams(CryptoDataParamsOIDOptional, GradeableVulnerabilities):
    digest_size = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(int)))
    hash_algorithm = attr.ib(validator=attr.validators.optional(attr.validators.instance_of((Hash, six.string_types))))

    @classmethod
    def get_gradeable_name(cls):
        return 'MAC'

    def __attrs_post_init__(self):
        if (self.digest_size is None) == (self.hash_algorithm is None):
            raise ValueError()

        if isinstance(self.hash_algorithm, six.string_types):
            object.__setattr__(self, 'hash_algorithm', Hash[self.hash_algorithm])

        if self.digest_size is None:
            object.__setattr__(self, 'digest_size', self.hash_algorithm.value.digest_size)

        attr.validate(self)


MAC = CryptoDataEnumOIDBase('MAC', CryptoDataEnumOIDBase.get_json_records(MACParams))


@attr.s(frozen=True)
class MACModeParams(CryptoDataParamsEnumString, GradeableVulnerabilities):
    name = attr.ib(validator=attr.validators.instance_of(six.string_types))

    @classmethod
    def get_gradeable_name(cls):
        return 'MAC mode'


class MACMode(enum.Enum):
    ENCRYPT_THEN_MAC = MACModeParams(
        code='encrypt_then_mac',
        name='encrypt then MAC',
        vulnerabilities=[],
    )
    ENCRYPT_AND_MAC = MACModeParams(
        code='encrypt_and_mac',
        name='encrypt and MAC',
        vulnerabilities=[
            Vulnerability(attack_type=AttackType.FORGERY_ATTACK, grade=Grade.WEAK, named=None),
        ],
    )
    MAC_THEN_ENCRYP = MACModeParams(
        code='mac_then_encrypt',
        name='MAC then encrypt',
        vulnerabilities=[],
    )


@attr.s(frozen=True)
class SignatureParams(CryptoDataParamsOIDOptional, GradeableComplex):
    key_type = attr.ib(
        converter=convert_enum(Authentication),
        validator=attr.validators.instance_of(Authentication),
    )
    hash_algorithm = attr.ib(
        converter=convert_enum(Hash),
        validator=attr.validators.optional(attr.validators.instance_of(Hash)),
    )

    def __attrs_post_init__(self):
        object.__setattr__(self, 'gradeables', [self.hash_algorithm.value])

        attr.validate(self)


Signature = CryptoDataEnumOIDBase('Signature', CryptoDataEnumOIDBase.get_json_records(SignatureParams))



# -*- coding: utf-8 -*-

import enum

import attr

from cryptodatahub.common.types import CryptoDataEnumBase, CryptoDataParamsNamed, convert_enum, convert_iterable


class EntityType(enum.Enum):
    FOR_PROFIT_ORG = 'for-profit organization'
    GOV_ORG = 'governmental organization'
    NONPROFIT_ORG = 'nonprofit organization'
    NOT_FOR_PROFIT_ORG = 'not-for-profit organization'
    PRIV_PERSON = 'private person'


class EntityRole(enum.Enum):
    CA_TRUST_STORE_OWNER = 'CA trust store owner'
    CLIENT_DEVELOPER = 'client developer'
    CT_LOG_OPERATOR = 'certificate transparency log operator'
    SERVER_DEVELOPER = 'server developer'
    STANDARD_DEVELOPER = 'standard developer'


@attr.s(frozen=True)
class EntityParams(CryptoDataParamsNamed):
    type = attr.ib(
        converter=convert_enum(EntityType),
        validator=attr.validators.instance_of(EntityType)
    )
    activities = attr.ib(
        converter=convert_iterable(convert_enum(EntityRole)),
        validator=attr.validators.deep_iterable(attr.validators.instance_of(EntityRole))
    )


class EntityBase(CryptoDataEnumBase):
    @classmethod
    def get_items_by_role(cls, role):
        if not hasattr(cls, '_ITEMS_BY_ROLE'):
            items_by_role = {}

            for activity in EntityRole:
                items_by_role[activity] = []

            for entity in cls:
                for activity in entity.value.activities:
                    items_by_role[activity].append(entity)

            cls._ITEMS_BY_ROLE = {
                role: tuple(activities)
                for role, activities in items_by_role.items()
            }

        return cls._ITEMS_BY_ROLE[role]


Entity = EntityBase('Entity', EntityBase.get_json_records(EntityParams))


class ClientType(enum.Enum):
    CT_LOG_OPERATOR = 'certificate transparency log operator'
    WEB_BROWSER = 'web browser'


@attr.s(frozen=True)
class ClientParams(CryptoDataParamsNamed):
    type = attr.ib(
        converter=convert_enum(ClientType),
        validator=attr.validators.instance_of(ClientType)
    )
    developer = attr.ib(
        converter=convert_enum(Entity),
        validator=attr.validators.instance_of(Entity)
    )

    def __str__(self):
        return '{} {}'.format(self.developer.value.name, self.name)


Client = CryptoDataEnumBase('Client', CryptoDataEnumBase.get_json_records(ClientParams))


class ServerType(enum.Enum):
    FTP_SERVER = 'ftp server'
    MAIL_SERVER = 'mail server'
    WEB_SERVER = 'web server'
    TCP_SERVER = 'tcp server'


@attr.s(frozen=True)
class ServerParams(CryptoDataParamsNamed):
    types = attr.ib(
        converter=convert_iterable(convert_enum(ServerType)),
        validator=attr.validators.deep_iterable(attr.validators.instance_of(ServerType))
    )
    developers = attr.ib(
        converter=convert_iterable(convert_enum(Entity)),
        validator=attr.validators.deep_iterable(attr.validators.instance_of(Entity))
    )

    def __str__(self):
        server_str = self.name
        if self.developers:
            server_str += ' ({})'.format(','.join(map(lambda developer: developer.value.name, self.developers)))
        return server_str


Server = CryptoDataEnumBase('Server', CryptoDataEnumBase.get_json_records(ServerParams))



# -*- coding: utf-8 -*-

import abc
import base64
import codecs
import collections
import datetime
import enum
import json
import inspect
import os
import re
import unicodedata

import dateutil.parser
import six
from six.moves import collections_abc

try:
    import pathlib
except ImportError:  # pragma: no cover
    import pathlib2 as pathlib  # pragma: no cover

import attr
import urllib3

from cryptodatahub.common.exception import InvalidValue


class _ConverterBase(object):
    @abc.abstractmethod
    def __call__(self, value):
        raise NotImplementedError()

    @abc.abstractmethod
    def __repr__(self):
        raise NotImplementedError()


@attr.s(repr=False, slots=True, hash=True)
class _DictObjectConverter(_ConverterBase):
    object_type = attr.ib(validator=attr.validators.instance_of(type))

    def __call__(self, value):
        if value is None:
            return None

        try:
            return self.object_type(**value)
        except TypeError:
            pass

        return value

    def __repr__(self):
        return '<dict to object converter>'


def convert_dict_to_object(object_type):
    return _DictObjectConverter(object_type)


@attr.s(repr=False, slots=True, hash=True)
class _ValueToObjectConverter(_ConverterBase):
    object_type = attr.ib(validator=attr.validators.instance_of(type))
    value_converter = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(_ConverterBase)))

    def __call__(self, value):
        if value is None:
            return None

        if isinstance(value, self.object_type):
            return value

        if self.value_converter is not None:
            value = self.value_converter(value)

        return self.object_type(value)

    def __repr__(self):
        return '<value to object converter>'


def convert_value_to_object(object_type, value_converter=None):
    return _ValueToObjectConverter(object_type, value_converter)


@attr.s(repr=False, slots=True, hash=True)
class _EnumConverter(_ConverterBase):
    enum_type = attr.ib(validator=attr.validators.instance_of(type))

    def __call__(self, value):
        if value is None:
            return None

        if not isinstance(value, six.string_types):
            return value

        try:
            value = getattr(self.enum_type, value.upper())
        except AttributeError:
            pass

        return value

    def __repr__(self):
        return '<enum converter>'


def convert_enum(enum_type):
    return _EnumConverter(enum_type)


@attr.s(frozen=True)
class Base64Data(object):
    value = attr.ib(validator=attr.validators.instance_of((bytes, bytearray)))

    def _asdict(self):
        return str(self)

    def __str__(self):
        return base64.b64encode(self.value).decode('ascii')


@attr.s(repr=False, slots=True, hash=True)
class _DateTimeConverter(_ConverterBase):
    strptime_format = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(str)))

    def __call__(self, date_time):
        if date_time is None:
            return None

        if isinstance(date_time, datetime.datetime):
            return date_time

        try:
            if self.strptime_format is None:
                date_time = dateutil.parser.isoparse(date_time)
            else:
                date_time = datetime.datetime.strptime(date_time, self.strptime_format)
        except (TypeError, ValueError):
            pass

        return date_time

    def __repr__(self):
        return '<datetime converter>'


def convert_datetime(strptime_format=None):
    return _DateTimeConverter(strptime_format)


@attr.s(repr=False, slots=True, hash=True)
class _Base64DataConverter(_ConverterBase):
    def __call__(self, value):
        if value is None:
            return None

        if isinstance(value, bytearray) or (six.PY3 and isinstance(value, bytes)):
            return Base64Data(value)

        if not isinstance(value, six.string_types):
            return value

        try:
            value = Base64Data(base64.b64decode(value))
        except (ValueError, TypeError):
            pass

        return value

    def __repr__(self):
        return '<base64 data converter>'


def convert_base64_data():
    return _Base64DataConverter()


@attr.s(repr=False, slots=True, hash=True)
class _BigNumberConverter(_ConverterBase):
    def __call__(self, value):
        if value is None:
            return None

        if not isinstance(value, collections_abc.Iterable):
            return value

        if not all(map(lambda big_number_part: isinstance(big_number_part, six.string_types), value)):
            return value

        try:
            value = int(''.join(value).replace('0x', '').replace(' ', '').replace(',', ''), 16)
        except ValueError:
            pass

        return value

    def __repr__(self):
        return '<big number converter>'


def convert_big_enum():
    return _BigNumberConverter()


@attr.s(repr=False, slots=True, hash=True)
class _IterableConverter(_ConverterBase):
    member_converter = attr.ib(validator=attr.validators.instance_of(collections_abc.Callable))

    def __call__(self, iterable):
        if iterable is None:
            return None

        try:
            for idx, member in enumerate(iterable):
                iterable[idx] = self.member_converter(member)
        except (TypeError, ValueError):
            pass

        return iterable

    def __repr__(self):
        return '<iterable converter>'


def convert_iterable(member_converter):
    return _IterableConverter(member_converter)


@attr.s(repr=False, slots=True, hash=True)
class _MappingConverter(_ConverterBase):
    key_converter = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(collections_abc.Callable)))
    value_converter = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(collections_abc.Callable)))

    def __call__(self, mapping):
        if mapping is None:
            return None

        if not isinstance(mapping, collections_abc.Mapping):
            return mapping

        try:
            key_value_pairs = [[key, value] for key, value in mapping.items()]
            if self.key_converter is not None:
                for pair in key_value_pairs:
                    pair[0] = self.key_converter(pair[0])
            if self.value_converter is not None:
                for pair in key_value_pairs:
                    pair[1] = self.value_converter(pair[1])
            mapping = type(mapping)(key_value_pairs)
        except (TypeError, ValueError):
            pass

        return mapping

    def __repr__(self):
        return '<mapping converter>'


def convert_mapping(key_converter=None, value_converter=None):
    return _MappingConverter(key_converter, value_converter)


@attr.s(frozen=True)
class ClientVersion(object):
    parts = attr.ib(validator=attr.validators.deep_iterable(attr.validators.instance_of(int)))

    @classmethod
    def from_str(cls, version_str):
        try:
            return cls((int(version_str), ))
        except ValueError:
            pass

        return cls(tuple(map(int, version_str.split('.'))))

    def __str__(self):
        return '.'.join(map(str, self.parts))


@attr.s(repr=False, slots=True, hash=True)
class _ClientVersionConverter(_ConverterBase):
    def __call__(self, version):
        if version is None:
            return None

        if not isinstance(version, six.string_types):
            return version

        try:
            version = ClientVersion.from_str(version)
        except (TypeError, ValueError):
            pass

        return version

    def __repr__(self):
        return '<client version converter>'


def convert_client_version():
    return _ClientVersionConverter()


@attr.s(repr=False, slots=True, hash=True)
class _UrlConverter(_ConverterBase):
    def __call__(self, value):
        if value is None:
            return None

        if not isinstance(value, six.string_types):
            return value

        try:
            value = urllib3.util.parse_url(value)
        except urllib3.exceptions.LocationParseError:
            pass

        return value

    def __repr__(self):
        return '<url converter>'


def convert_url():
    return _UrlConverter()


@attr.s(repr=False, slots=True, hash=True)
class _VariadicConverter(_ConverterBase):
    converters = attr.ib(
        validator=attr.validators.deep_iterable(attr.validators.instance_of(_ConverterBase))
    )

    def __call__(self, convertable):
        if convertable is None:
            return None

        for converter in self.converters:
            converted = converter(convertable)
            if id(converted) != id(convertable):
                return converted

        return convertable

    def __repr__(self):
        return '<variadic converter>'


def convert_variadic(converters):
    return _VariadicConverter(converters)


class CryptoDataParamsBase(object):
    @classmethod
    def get_init_attribute_names(cls):
        return [
            six.ensure_text(name)
            for name, attribute in attr.fields_dict(cls).items()
            if attribute.init
        ]

    def _asdict_filter(self, attribute, value):
        if attribute.name.startswith('_'):
            return False

        if attribute.default != attr.NOTHING and value == attribute.default:
            return False

        return True

    def _asdict_serializer(self, _, __, value):
        if hasattr(value, '_asdict'):
            return getattr(value, '_asdict')()
        if isinstance(value, enum.Enum):
            return value.name

        return value

    def _asdict(self):
        return attr.asdict(
            self,
            filter=self._asdict_filter,
            dict_factory=collections.OrderedDict,
            value_serializer=self._asdict_serializer
        )


@attr.s(frozen=True)
class CryptoDataParamsFetchedBase(CryptoDataParamsBase):
    @property
    @abc.abstractmethod
    def identifier(self):
        raise NotImplementedError()


@attr.s(frozen=True)
class CryptoDataParamsNamed(CryptoDataParamsBase):
    name = attr.ib(validator=attr.validators.instance_of(six.string_types))
    long_name = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(six.string_types)))

    def __str__(self):
        return self.name


@attr.s(frozen=True)
class CryptoDataParamsEnumNumeric(CryptoDataParamsBase):
    code = attr.ib()

    @code.validator
    def _validator_code(self, _, value):
        if not isinstance(value, int):
            raise ValueError(value)
        if value < 0:
            raise ValueError(value)
        if value >= 2 ** (self.get_code_size() * 8):
            raise ValueError(value)

    @classmethod
    def get_code_size(cls):
        raise NotImplementedError()


@attr.s(frozen=True)
class CryptoDataParamsEnumString(CryptoDataParamsBase):
    code = attr.ib(validator=attr.validators.instance_of(six.string_types))

    def __str__(self):
        return self.code

    def _asdict(self):
        return str(self)

    def get_code_size(self):
        return len(self.code)


@attr.s(frozen=True)
class CryptoDataParamsOIDOptional(CryptoDataParamsNamed):
    oid = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(six.string_types)))


class CryptoDataEnumBase(enum.Enum):
    @classmethod
    def get_json_records(cls, param_class):
        return collections.OrderedDict([
            (six.ensure_str(unicodedata.normalize('NFD', name).encode('ascii', 'ignore')), param_class(**{
                init_param_name.replace('-', '_'): value
                for init_param_name, value in params.items()
                if not init_param_name.startswith('_')
            }))
            for name, params in cls.get_json_object(param_class).items()
        ])

    @staticmethod
    def get_file_name_from_param_class(param_class):
        if not param_class.__name__.endswith('Params'):
            raise TypeError(param_class)

        enum_class_name = param_class.__name__[:-6]
        enum_class_name_parts = [
            name_part.lower()
            for name_part in re.split("([A-Z]+[^A-Z]*)", enum_class_name)
            if name_part
        ]

        return '-'.join(enum_class_name_parts) + '.json'

    @classmethod
    def get_json_path(cls, param_class):
        return (
            pathlib.Path(inspect.getfile(param_class)).parent /
            cls.get_file_name_from_param_class(param_class)
        )

    @classmethod
    def get_json_encoding(cls):
        return 'ascii'

    @classmethod
    def is_json_encoding_ascii(cls):
        return cls.get_json_encoding() == 'ascii'

    @classmethod
    def get_json_object(cls, param_class):
        json_path = cls.get_json_path(param_class)
        with codecs.open(str(json_path), 'r', encoding=cls.get_json_encoding()) as json_file:
            return json.load(json_file, object_pairs_hook=collections.OrderedDict)

    @classmethod
    def dump_json(cls, json_object):
        return json.dumps(json_object, ensure_ascii=cls.is_json_encoding_ascii(), indent=4) + os.linesep

    @classmethod
    def set_json(cls, param_class, json_object):
        json_path = cls.get_json_path(param_class)
        with codecs.open(str(json_path), 'w+', encoding=cls.get_json_encoding()) as json_file:
            json_file.write(cls.dump_json(json_object))

    @classmethod
    def get_json(cls, param_class):
        return cls.get_json_object(param_class)

    @classmethod
    def _from_attr(cls, attr_name, value):
        for item in cls:
            if value == getattr(item.value, attr_name):
                return item

        raise InvalidValue(value, cls, attr_name)


class CryptoDataEnumCodedBase(CryptoDataEnumBase):
    @classmethod
    def from_code(cls, code):
        return cls._from_attr('code', code)


class CryptoDataEnumOIDBase(CryptoDataEnumCodedBase):
    @classmethod
    def from_oid(cls, oid):
        return cls._from_attr('oid', oid)



# -*- coding: utf-8 -*-

import collections
import enum
import six

import attr

from cryptodatahub.common.entity import Client
from cryptodatahub.common.types import (
    ClientVersion,
    CryptoDataEnumBase,
    CryptoDataParamsBase,
    convert_client_version,
    convert_dict_to_object,
    convert_enum,
    convert_iterable,
)

from cryptodatahub.tls.algorithm import (
    TlsCertificateCompressionAlgorithm,
    TlsCipherSuite,
    TlsCompressionMethod,
    TlsECPointFormat,
    TlsExtensionType,
    TlsNamedCurve,
    TlsProtocolName,
    TlsPskKeyExchangeMode,
    TlsSignatureAndHashAlgorithm,
    TlsTokenBindingParamater,
)
from cryptodatahub.tls.version import TlsVersion


@attr.s(frozen=True)
class ClientTokenBindingParams(CryptoDataParamsBase):
    parameters = attr.ib(
        converter=convert_iterable(convert_enum(TlsTokenBindingParamater)),
        validator=attr.validators.deep_iterable(attr.validators.instance_of(TlsTokenBindingParamater))
    )
    protocol_version = attr.ib(validator=attr.validators.instance_of(six.string_types))


@attr.s(frozen=True)
class ClientGreaseParams(CryptoDataParamsBase):
    cipher_suites = attr.ib(default=False, validator=attr.validators.instance_of(bool))
    extension_types = attr.ib(default=False, validator=attr.validators.instance_of(bool))
    extensions = attr.ib(
        default=[],
        converter=convert_iterable(convert_enum(TlsExtensionType)),
        validator=attr.validators.deep_iterable(attr.validators.instance_of(TlsExtensionType))
    )


class ClientConfigurationChange(enum.IntEnum):
    pass


@attr.s(frozen=True)
class ClientVersionedParamsBase(CryptoDataParamsBase):
    client = attr.ib(converter=convert_enum(Client), validator=attr.validators.instance_of(Client))
    first_version = attr.ib(converter=convert_client_version(), validator=attr.validators.instance_of(ClientVersion))
    last_version = attr.ib(converter=convert_client_version(), validator=attr.validators.instance_of(ClientVersion))

    def __str__(self):
        return '{} ({} - {})'.format(
            self.client.value, self.first_version, self.last_version
        )


@attr.s(frozen=True)
class ClientVersionedParams(ClientVersionedParamsBase):
    changes = attr.ib(
        converter=frozenset,
        validator=attr.validators.deep_iterable(attr.validators.instance_of(ClientConfigurationChange))
    )


@attr.s(frozen=True)
class ClientExtensionParams(CryptoDataParamsBase):  # pylint: disable=too-many-instance-attributes
    application_layer_protocol_negotiation = attr.ib(
        default=None,
        converter=convert_iterable(convert_enum(TlsProtocolName)),
        validator=attr.validators.optional(
            attr.validators.deep_iterable(attr.validators.instance_of(TlsProtocolName))
        )
    )
    application_layer_protocol_settings = attr.ib(
        default=None,
        converter=convert_iterable(convert_enum(TlsProtocolName)),
        validator=attr.validators.optional(
            attr.validators.deep_iterable(attr.validators.instance_of(TlsProtocolName))
        )
    )
    compress_certificate = attr.ib(
        default=None,
        converter=convert_iterable(convert_enum(TlsCertificateCompressionAlgorithm)),
        validator=attr.validators.optional(
            attr.validators.deep_iterable(attr.validators.instance_of(TlsCertificateCompressionAlgorithm))
        )
    )
    delegated_credentials = attr.ib(
        default=None,
        converter=convert_iterable(convert_enum(TlsSignatureAndHashAlgorithm)),
        validator=attr.validators.optional(
            attr.validators.deep_iterable(attr.validators.instance_of(TlsSignatureAndHashAlgorithm))
        )
    )
    ec_point_formats = attr.ib(
        default=None,
        converter=convert_iterable(convert_enum(TlsECPointFormat)),
        validator=attr.validators.optional(
            attr.validators.deep_iterable(attr.validators.instance_of(TlsECPointFormat))
        )
    )
    key_share = attr.ib(
        default=None,
        converter=convert_iterable(convert_enum(TlsNamedCurve)),
        validator=attr.validators.optional(
            attr.validators.deep_iterable(attr.validators.instance_of(TlsNamedCurve))
        )
    )
    key_share_reserved = attr.ib(
        default=None,
        converter=convert_iterable(convert_enum(TlsNamedCurve)),
        validator=attr.validators.optional(
            attr.validators.deep_iterable(attr.validators.instance_of(TlsNamedCurve))
        )
    )
    post_handshake_auth = attr.ib(default=None)
    psk_key_exchange_modes = attr.ib(
        default=None,
        converter=convert_iterable(convert_enum(TlsPskKeyExchangeMode)),
        validator=attr.validators.optional(
            attr.validators.deep_iterable(attr.validators.instance_of(TlsPskKeyExchangeMode))
        )
    )
    record_size_limit = attr.ib(default=None, validator=attr.validators.optional(attr.validators.instance_of(int)))
    signature_algorithms = attr.ib(
        default=None,
        converter=convert_iterable(convert_enum(TlsSignatureAndHashAlgorithm)),
        validator=attr.validators.optional(
            attr.validators.deep_iterable(attr.validators.instance_of(TlsSignatureAndHashAlgorithm))
        )
    )
    supported_groups = attr.ib(
        default=None,
        converter=convert_iterable(convert_enum(TlsNamedCurve)),
        validator=attr.validators.optional(
            attr.validators.deep_iterable(attr.validators.instance_of(TlsNamedCurve))
        )
    )
    supported_versions = attr.ib(
        default=None,
        converter=convert_iterable(convert_enum(TlsVersion)),
        validator=attr.validators.optional(
            attr.validators.deep_iterable(attr.validators.instance_of(TlsVersion))
        )
    )
    token_binding = attr.ib(
        default=None,
        converter=convert_dict_to_object(ClientTokenBindingParams),
        validator=attr.validators.optional(attr.validators.instance_of(ClientTokenBindingParams))
    )


@attr.s(frozen=True)
class ClientCapabilities(CryptoDataParamsBase):  # pylint: disable=too-many-instance-attributes
    tls_versions = attr.ib(
        converter=convert_iterable(convert_enum(TlsVersion)),
        validator=attr.validators.deep_iterable(attr.validators.instance_of(TlsVersion))
    )
    cipher_suites = attr.ib(
        converter=convert_iterable(convert_enum(TlsCipherSuite)),
        validator=attr.validators.deep_iterable(attr.validators.instance_of(TlsCipherSuite))
    )
    compression_methods = attr.ib(
        converter=convert_iterable(convert_enum(TlsCompressionMethod)),
        validator=attr.validators.deep_iterable(attr.validators.instance_of(TlsCompressionMethod))
    )
    fallback_scsv = attr.ib(validator=attr.validators.instance_of(bool))
    empty_renegotiation_info_scsv = attr.ib(validator=attr.validators.instance_of(bool))
    grease = attr.ib(
        converter=convert_dict_to_object(ClientGreaseParams),
        validator=attr.validators.instance_of(ClientGreaseParams)
    )
    extension_types = attr.ib(
        converter=convert_iterable(convert_enum(TlsExtensionType)),
        validator=attr.validators.deep_iterable(attr.validators.instance_of(TlsExtensionType))
    )
    extension_params = attr.ib(
        converter=convert_dict_to_object(ClientExtensionParams),
        validator=attr.validators.optional(attr.validators.instance_of(ClientExtensionParams))
    )

    def _asdict(self):
        capabilities_dict = attr.asdict(self, dict_factory=collections.OrderedDict)

        capabilities_dict['extension_params'] = collections.OrderedDict([
            (param_name, param_value)
            for param_name, param_value in capabilities_dict['extension_params'].items()
            if param_value is not None
        ])

        return capabilities_dict


@attr.s(frozen=True)
class ClientParams(CryptoDataParamsBase):  # pylint: disable=too-many-instance-attributes
    meta = attr.ib(
        converter=convert_dict_to_object(ClientVersionedParams),
        validator=attr.validators.instance_of(ClientVersionedParams)
    )
    capabilities = attr.ib(
        converter=convert_dict_to_object(ClientCapabilities),
        validator=attr.validators.instance_of(ClientCapabilities)
    )


TlsClient = CryptoDataEnumBase('TlsClient', CryptoDataEnumBase.get_json_records(ClientParams))



# -*- coding: utf-8 -*-

import abc
import enum
import attr

import six

from cryptodatahub.common.algorithm import (
    BlockCipher,
    BlockCipherMode,
    Hash,
    KeyExchange,
    MAC,
    MACMode,
    NamedGroup,
    Signature,
)
from cryptodatahub.common.grade import GradeableComplex, GradeableVulnerabilities
from cryptodatahub.common.key import PublicKeySize
from cryptodatahub.common.parameter import DHParamWellKnown
from cryptodatahub.common.types import (
    CryptoDataEnumCodedBase,
    CryptoDataParamsEnumString,
    convert_enum,
    convert_variadic,
)


@attr.s(frozen=True)
class SshAlgorithmParams(CryptoDataParamsEnumString, GradeableComplex):
    @property
    @abc.abstractmethod
    def _gradeable_algorithms(self):
        raise NotImplementedError()

    def __attrs_post_init__(self):
        gradeables = []
        for algorithm in self._gradeable_algorithms:
            if isinstance(algorithm, six.string_types):
                gradeable = getattr(self, algorithm)
                if gradeable is not None:
                    gradeable = gradeable.value
            else:
                gradeable = algorithm

            if gradeable is not None:
                gradeables.append(gradeable)

        object.__setattr__(self, 'gradeables', gradeables)

        attr.validate(self)


@attr.s(frozen=True)
class EncryptionAlgorithmParams(SshAlgorithmParams):
    cipher = attr.ib(
        converter=convert_enum(BlockCipher),
        validator=attr.validators.optional(attr.validators.instance_of(BlockCipher))
    )
    mode = attr.ib(
        converter=convert_enum(BlockCipherMode),
        validator=attr.validators.optional(attr.validators.instance_of(BlockCipherMode))
    )

    @property
    def _gradeable_algorithms(self):
        return ('cipher', 'mode')


@attr.s(frozen=True)
class MacAlgorithmParams(SshAlgorithmParams):
    truncated_size = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(int)))
    mac = attr.ib(
        converter=convert_enum(MAC),
        validator=attr.validators.optional(attr.validators.instance_of(MAC))
    )
    mode = attr.ib(
        converter=convert_enum(MACMode),
        validator=attr.validators.optional(attr.validators.instance_of(MACMode))
    )

    @property
    def size(self):
        if self.truncated_size is not None:
            return self.truncated_size

        return self.mac.value.digest_size

    @property
    def _gradeable_algorithms(self):
        return ('mac', 'mode')


@attr.s(frozen=True)
class KexAlgorithmParams(SshAlgorithmParams):
    kex = attr.ib(
        converter=convert_enum(KeyExchange),
        validator=attr.validators.optional(attr.validators.instance_of(KeyExchange))
    )
    key_parameter = attr.ib(
        converter=convert_variadic((convert_enum(NamedGroup), convert_enum(DHParamWellKnown))),
        validator=attr.validators.optional(
            attr.validators.instance_of((NamedGroup, DHParamWellKnown, six.string_types))
        )
    )
    exchange_hash = attr.ib(
        converter=convert_enum(Hash),
        validator=attr.validators.optional(attr.validators.instance_of(Hash))
    )
    key_size = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(int)))

    def __attrs_post_init__(self):
        super(KexAlgorithmParams, self).__attrs_post_init__()

        if self.key_size is not None:
            gradeables = PublicKeySize(self.kex, self.key_size).gradeables
            if gradeables is None:
                self.gradeables.append(gradeables)
            else:
                self.gradeables.extend(gradeables)

        attr.validate(self)

    @property
    def _gradeable_algorithms(self):
        gradeables = ['kex', 'exchange_hash']

        if isinstance(self.key_parameter, DHParamWellKnown):
            gradeables.append(self.key_parameter.value)

        return gradeables


class SshHostKeyType(enum.Enum):
    HOST_KEY = 'host key'
    HOST_CERTIFICATE = 'host certificate'
    PGP_KEY = 'PGP key'
    SECURE_KEY = 'secure key'
    SECURE_CERTIFICATE = 'secure certificate'
    SPKI_KEY = 'SPKI key'
    X509_CERTIFICATE = 'X.509 certificate'
    X509_CERTIFICATE_CHAIN = 'X.509 certificate chain'


@attr.s(frozen=True)
class HostKeyAlgorithmParams(SshAlgorithmParams):
    key_type = attr.ib(
        converter=convert_enum(SshHostKeyType),
        validator=attr.validators.instance_of(SshHostKeyType)
    )
    signature = attr.ib(
        converter=convert_enum(Signature),
        validator=attr.validators.optional(attr.validators.instance_of(Signature))
    )

    @property
    def _gradeable_algorithms(self):
        return ('signature',)


@attr.s(frozen=True)
class CompressionAlgorithmParams(CryptoDataParamsEnumString, GradeableVulnerabilities):
    @classmethod
    def get_gradeable_name(cls):
        return 'compression'


@attr.s(frozen=True)
class EllipticCurveIdentifierParams(CryptoDataParamsEnumString):
    named_group = attr.ib(
        converter=convert_enum(NamedGroup),
        validator=attr.validators.instance_of(NamedGroup)
    )


SshEncryptionAlgorithm = CryptoDataEnumCodedBase(
    'SshEncryptionAlgorithm', CryptoDataEnumCodedBase.get_json_records(EncryptionAlgorithmParams)
)
SshMacAlgorithm = CryptoDataEnumCodedBase(
    'SshMacAlgorithm', CryptoDataEnumCodedBase.get_json_records(MacAlgorithmParams)
)
SshKexAlgorithm = CryptoDataEnumCodedBase(
    'SshKexAlgorithm', CryptoDataEnumCodedBase.get_json_records(KexAlgorithmParams)
)
SshHostKeyAlgorithm = CryptoDataEnumCodedBase(
    'SshHostKeyAlgorithm', CryptoDataEnumCodedBase.get_json_records(HostKeyAlgorithmParams)
)
SshCompressionAlgorithm = CryptoDataEnumCodedBase(
    'SshCompressionAlgorithm', CryptoDataEnumCodedBase.get_json_records(CompressionAlgorithmParams)
)
SshEllipticCurveIdentifier = CryptoDataEnumCodedBase(
    'SshEllipticCurveIdentifier', CryptoDataEnumCodedBase.get_json_records(EllipticCurveIdentifierParams)
)



# -*- coding: utf-8 -*-

import attr
import six

from cryptodatahub.common.algorithm import Hash, KeyExchange, Signature
from cryptodatahub.common.grade import GradeableComplex
from cryptodatahub.common.types import (
    CryptoDataEnumBase,
    CryptoDataEnumCodedBase,
    CryptoDataParamsEnumNumeric,
    convert_enum,
    convert_variadic,
)


@attr.s(frozen=True)
class AlgorithmParams(CryptoDataParamsEnumNumeric, GradeableComplex):
    name = attr.ib(validator=attr.validators.instance_of(six.string_types))
    zone_transfer = attr.ib(validator=attr.validators.instance_of(bool))
    transaction_security = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(bool)))
    algorithm = attr.ib(
        converter=convert_variadic((convert_enum(Signature), convert_enum(KeyExchange))),
        validator=attr.validators.optional(attr.validators.instance_of(CryptoDataEnumBase))
    )

    def __attrs_post_init__(self):
        if self.algorithm is None:
            object.__setattr__(self, 'gradeables', [])
        else:
            object.__setattr__(self, 'gradeables', [self.algorithm.value])

        attr.validate(self)

    @classmethod
    def get_code_size(cls):
        return 1

    def __str__(self):
        return self.name


DnsSecAlgorithm = CryptoDataEnumCodedBase('DnsSecAlgorithm', CryptoDataEnumCodedBase.get_json_records(AlgorithmParams))


@attr.s(frozen=True)
class DigestTypeParams(CryptoDataParamsEnumNumeric, GradeableComplex):
    name = attr.ib(validator=attr.validators.instance_of(six.string_types))
    hash = attr.ib(
        converter=convert_enum(Hash),
        validator=attr.validators.instance_of(Hash),
    )

    def __attrs_post_init__(self):
        object.__setattr__(self, 'gradeables', [self.hash.value])

        attr.validate(self)

    @classmethod
    def get_code_size(cls):
        return 1

    def __str__(self):
        return self.name


DnsSecDigestType = CryptoDataEnumCodedBase(
    'DnsSecDigestType', CryptoDataEnumCodedBase.get_json_records(DigestTypeParams)
)


@attr.s(frozen=True)
class RrTypeParams(CryptoDataParamsEnumNumeric):
    name = attr.ib(validator=attr.validators.instance_of(six.string_types))
    description = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(six.string_types)))

    @classmethod
    def get_code_size(cls):
        return 2

    def __str__(self):
        return self.name


DnsRrType = CryptoDataEnumCodedBase(
    'DnsRrType', CryptoDataEnumCodedBase.get_json_records(RrTypeParams)
)



# -*- coding: utf-8 -*-

import enum
import six

import attr


@attr.s(init=False)
class InvalidValue(Exception):
    value = attr.ib()

    def __init__(self, value, type_class, class_member=None):
        if isinstance(value, enum.IntEnum):
            message = hex(value.value)
        elif isinstance(value, int):
            message = hex(value)
        else:
            message = value
        message = hex(value) if isinstance(value, int) else repr(value)
        type_name = type_class.__name__ if hasattr(type_class, '__name__') else str(type(type_class))
        message = six.ensure_text('{} is not a valid {}').format(message, type_name)
        if class_member is not None:
            message = six.ensure_text('{} {} value').format(message, class_member)

        super(InvalidValue, self).__init__(message)

        self.value = value



# -*- coding: utf-8 -*-

__title__ = 'CryptoDataHub'
__technical_name__ = __title__.lower()
__version__ = '0.12.6'
__description__ = 'Repository of cryptography-related data'
__author__ = 'Szilárd Pfeiffer'
__author_email__ = 'coroner@pfeifferszilard.hu'
__url__ = 'https://gitlab.com/coroner/' + __technical_name__
__license__ = 'MPL-2.0'



# -*- coding: utf-8 -*-

import abc
import enum
import functools

import six

import attr

from cryptodatahub.common.types import (
    CryptoDataEnumBase,
    CryptoDataEnumOIDBase,
    CryptoDataParamsNamed,
    convert_enum,
    convert_iterable,
    convert_dict_to_object,
)


@functools.total_ordering
@attr.s(frozen=True, order=False, eq=False, hash=True)
class GradeTypeParams(CryptoDataParamsNamed):
    _numeric = attr.ib(validator=attr.validators.instance_of(six.integer_types))

    def __eq__(self, other):
        return self._numeric == other._numeric

    def __lt__(self, other):
        return self._numeric < other._numeric


class Grade(enum.Enum):
    SECURE = GradeTypeParams(
        name='secure',
        long_name=None,
        numeric=0,
    )
    DEPRECATED = GradeTypeParams(
        name='deprecated',
        long_name=None,
        numeric=-1,
    )
    WEAK = GradeTypeParams(
        name='weak',
        long_name=None,
        numeric=-2,
    )
    INSECURE = GradeTypeParams(
        name='insecure',
        long_name=None,
        numeric=-3,
    )


@attr.s(frozen=True)
class AttackTypeParams(CryptoDataParamsNamed):
    pass


AttackType = CryptoDataEnumOIDBase('AttackType', CryptoDataEnumBase.get_json_records(AttackTypeParams))


@attr.s(frozen=True)
class AttackNamedParams(CryptoDataParamsNamed):
    grade = attr.ib(
        converter=convert_enum(Grade),
        validator=attr.validators.instance_of(Grade)
    )
    attack_type = attr.ib(
        converter=convert_enum(AttackType),
        validator=attr.validators.optional(attr.validators.instance_of(AttackType))
    )


AttackNamed = CryptoDataEnumOIDBase('AttackNamed', CryptoDataEnumBase.get_json_records(AttackNamedParams))


@attr.s(frozen=True, eq=False)
class Gradeable(object):
    @staticmethod
    def _get_vulnerbilities(obj):
        if obj is None:
            return [None]

        if isinstance(obj, Vulnerability):
            return [obj]

        if isinstance(obj, (list, tuple)):
            result = []
            for gradeable in obj:
                if isinstance(gradeable, GradeableComplex):
                    result.append(Gradeable._get_vulnerbilities(gradeable.gradeables))
                elif isinstance(gradeable, GradeableVulnerabilities):
                    result.append(gradeable.vulnerabilities)
                else:
                    result.append(Gradeable._get_vulnerbilities(gradeable))

            return result

        raise NotImplementedError(type(obj))

    @staticmethod
    def _flatten_vulnerabilities(vulnerabilities):
        if not vulnerabilities:
            return vulnerabilities

        if isinstance(vulnerabilities[0], list):
            return (
                Gradeable._flatten_vulnerabilities(vulnerabilities[0]) +
                Gradeable._flatten_vulnerabilities(vulnerabilities[1:])
            )
        return vulnerabilities[:1] + Gradeable._flatten_vulnerabilities(vulnerabilities[1:])

    @staticmethod
    def get_min_grade(vulnerabilities):
        vulnerabilities = Gradeable._flatten_vulnerabilities((Gradeable._get_vulnerbilities(vulnerabilities)))
        if not vulnerabilities:
            return Grade.SECURE

        grades = set(
            vulnerability.grade
            for vulnerability in vulnerabilities
            if vulnerability is not None
        )
        if grades:
            return min(grades, key=lambda grade: grade.value)

        return None

    @property
    @abc.abstractmethod
    def min_grade(self):
        raise NotImplementedError()


@attr.s(frozen=True)
class Vulnerability(object):
    attack_type = attr.ib(
        converter=convert_enum(AttackType),
        validator=attr.validators.optional(attr.validators.instance_of(AttackType))
    )
    grade = attr.ib(
        converter=convert_enum(Grade),
        validator=attr.validators.instance_of(Grade)
    )
    named = attr.ib(
        converter=convert_enum(AttackNamed),
        validator=attr.validators.optional(attr.validators.instance_of(AttackNamed))
    )


@attr.s(frozen=True, eq=False)
class GradeableSimple(Gradeable):
    @property
    @abc.abstractmethod
    def grade(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def __str__(self):
        raise NotImplementedError()

    @property
    def min_grade(self):
        return self.grade


@attr.s(frozen=True, eq=False)
class GradeableVulnerabilities(Gradeable):
    vulnerabilities = attr.ib(
        converter=convert_iterable(convert_dict_to_object(Vulnerability)),
        validator=attr.validators.optional(
            validator=attr.validators.deep_iterable(member_validator=attr.validators.instance_of(Vulnerability))
        )
    )

    @classmethod
    @abc.abstractmethod
    def get_gradeable_name(cls):
        raise NotImplementedError()

    @property
    def min_grade(self):
        return self.get_min_grade(self.vulnerabilities)


@attr.s(eq=False)
class GradeableComplex(Gradeable):
    gradeables = attr.ib(
        init=False,
        default=None,
        validator=attr.validators.optional(attr.validators.deep_iterable(
            member_validator=attr.validators.optional(attr.validators.instance_of(Gradeable))
        ))
    )

    @property
    def min_grade(self):
        return self.get_min_grade(self.gradeables)



# -*- coding: utf-8 -*-

import abc
import collections
import six

import attr

from cryptodatahub.common.algorithm import (
    Authentication,
    BlockCipher,
    BlockCipherMode,
    Hash,
    KeyExchange,
    MAC,
    NamedGroup,
)
from cryptodatahub.common.grade import GradeableComplex
from cryptodatahub.common.types import (
    CryptoDataEnumCodedBase,
    CryptoDataParamsEnumNumeric,
    CryptoDataParamsEnumString,
    CryptoDataParamsNamed,
    convert_enum,
)

from cryptodatahub.tls.version import TlsVersion


@attr.s(frozen=True)
class CompressionMethodParams(CryptoDataParamsNamed, CryptoDataParamsEnumNumeric):
    @classmethod
    def get_code_size(cls):
        return 1


TlsCompressionMethod = CryptoDataEnumCodedBase(
    'TlsCompressionMethod', CryptoDataEnumCodedBase.get_json_records(CompressionMethodParams)
)


@attr.s(frozen=True)
class NamedCurveParams(CryptoDataParamsEnumNumeric, GradeableComplex):
    named_group = attr.ib(
        converter=convert_enum(NamedGroup),
        validator=attr.validators.optional(attr.validators.instance_of(NamedGroup)),
    )

    def __attrs_post_init__(self):
        if self.named_group is not None:
            object.__setattr__(self, 'gradeables', [self.named_group.value])

        attr.validate(self)

    def __str__(self):
        return str(self.named_group.value)

    @classmethod
    def get_code_size(cls):
        return 2


TlsNamedCurve = CryptoDataEnumCodedBase('TlsNamedCurve', CryptoDataEnumCodedBase.get_json_records(NamedCurveParams))


@attr.s(frozen=True)
class HashAndSignatureAlgorithmParams(CryptoDataParamsEnumNumeric, GradeableComplex):
    hash_algorithm = attr.ib(
        converter=convert_enum(Hash),
        validator=attr.validators.optional(attr.validators.instance_of(Hash)),
    )
    signature_algorithm = attr.ib(
        converter=convert_enum(Authentication),
        validator=attr.validators.optional(attr.validators.instance_of(Authentication)),
    )

    def __attrs_post_init__(self):
        vulnerabilities = []
        if self.hash_algorithm is not None:
            vulnerabilities.append(self.hash_algorithm.value)

        object.__setattr__(self, 'gradeables', vulnerabilities)

        attr.validate(self)

    def __str__(self):
        if self.hash_algorithm:
            hash_algorithm = self.hash_algorithm.value
        else:
            hash_algorithm = 'none'

        if self.signature_algorithm == Authentication.ANONYMOUS:
            signature_algorithm = 'no'
        else:
            signature_algorithm = self.signature_algorithm.value

        return '{} with {} encryption'.format(hash_algorithm, signature_algorithm)

    @classmethod
    def get_code_size(cls):
        return 2


TlsSignatureAndHashAlgorithm = CryptoDataEnumCodedBase(
    'TlsSignatureAndHashAlgorithm', CryptoDataEnumCodedBase.get_json_records(HashAndSignatureAlgorithmParams)
)


@attr.s(frozen=True)
class EcPointFormatParams(CryptoDataParamsNamed, CryptoDataParamsEnumNumeric):
    @classmethod
    def get_code_size(cls):
        return 1


TlsECPointFormat = CryptoDataEnumCodedBase(
    'TlsECPointFormat', CryptoDataEnumCodedBase.get_json_records(EcPointFormatParams)
)


@attr.s(frozen=True)
class ProtocolNameParams(CryptoDataParamsEnumString):
    pass


TlsProtocolName = CryptoDataEnumCodedBase(
    'TlsProtocolName', CryptoDataEnumCodedBase.get_json_records(ProtocolNameParams)
)


@attr.s(frozen=True)
class NextProtocolNameParams(CryptoDataParamsEnumString):
    pass


TlsNextProtocolName = CryptoDataEnumCodedBase(
    'TlsNextProtocolName', CryptoDataEnumCodedBase.get_json_records(NextProtocolNameParams)
)


@attr.s(frozen=True)
class ExtensionTypeParams(CryptoDataParamsEnumNumeric):
    @classmethod
    def get_code_size(cls):
        return 2


TlsExtensionType = CryptoDataEnumCodedBase(
    'ExtensionType', CryptoDataEnumCodedBase.get_json_records(ExtensionTypeParams)
)


@attr.s(frozen=True)
class TokenBindingParamaterParams(CryptoDataParamsEnumNumeric):
    @classmethod
    def get_code_size(cls):
        return 1


TlsTokenBindingParamater = CryptoDataEnumCodedBase(
    'TlsTokenBindingParamater', CryptoDataEnumCodedBase.get_json_records(TokenBindingParamaterParams)
)


@attr.s(frozen=True)
class PskKeyExchangeModeParams(CryptoDataParamsEnumNumeric):
    @classmethod
    def get_code_size(cls):
        return 1


TlsPskKeyExchangeMode = CryptoDataEnumCodedBase(
    'TlsPskKeyExchangeMode', CryptoDataEnumCodedBase.get_json_records(PskKeyExchangeModeParams)
)


@attr.s(frozen=True)
class CertificateCompressionAlgorithmParams(CryptoDataParamsEnumNumeric):
    @classmethod
    def get_code_size(cls):
        return 2


TlsCertificateCompressionAlgorithm = CryptoDataEnumCodedBase(
    'TlsCertificateCompressionAlgorithm',
    CryptoDataEnumCodedBase.get_json_records(CertificateCompressionAlgorithmParams)
)


@attr.s(frozen=True)
class CipherParamsBase(CryptoDataParamsEnumNumeric, GradeableComplex):  # pylint: disable=too-many-instance-attributes
    iana_name = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(six.string_types)))
    iana_name = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(six.string_types)))
    openssl_name = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(six.string_types)))
    key_exchange = attr.ib(
        converter=convert_enum(KeyExchange),
        validator=attr.validators.optional(attr.validators.instance_of(KeyExchange)),
    )
    authentication = attr.ib(
        converter=convert_enum(Authentication),
        validator=attr.validators.optional(attr.validators.instance_of(Authentication)),
    )
    bulk_cipher = attr.ib(
        converter=convert_enum(BlockCipher),
        validator=attr.validators.optional(attr.validators.instance_of(BlockCipher)),
    )
    block_cipher_mode = attr.ib(
        converter=convert_enum(BlockCipherMode),
        validator=attr.validators.optional(attr.validators.instance_of(BlockCipherMode)),
    )
    mac = attr.ib(
        converter=convert_enum(MAC),
        validator=attr.validators.optional(attr.validators.instance_of(MAC)),
    )
    authenticated_encryption = attr.ib(validator=attr.validators.instance_of(bool))
    initial_version = attr.ib(
        converter=convert_enum(TlsVersion),
        validator=attr.validators.instance_of(TlsVersion),
    )
    last_version = attr.ib(init=False, validator=attr.validators.instance_of(TlsVersion))
    export_grade = attr.ib(init=False, validator=attr.validators.instance_of(bool))

    @classmethod
    @abc.abstractmethod
    def get_code_size(cls):
        raise NotImplementedError()

    def __attrs_post_init__(self):
        if self.initial_version == TlsVersion.SSL2:
            object.__setattr__(self, 'last_version', TlsVersion.SSL2)
        else:
            if self.code & 0xff00 in [0x1300, 0x7e00, 0x7f00]:
                object.__setattr__(self, 'last_version', TlsVersion.TLS1_3)
            else:
                object.__setattr__(self, 'last_version', TlsVersion.TLS1_2)

        object.__setattr__(self, 'export_grade', self.iana_name is not None and '_EXPORT' in self.iana_name)

        vulnerability_parts = collections.OrderedDict([
            (KeyExchange, self.key_exchange),
            (Authentication, self.authentication),
            (BlockCipher, self.bulk_cipher),
            (BlockCipherMode, self.block_cipher_mode),
            (MAC, self.mac),
        ])

        gradeables = []
        for algorithm in vulnerability_parts.values():
            if algorithm is None:
                continue

            gradeables.append(algorithm.value)

        object.__setattr__(self, 'gradeables', gradeables)

        attr.validate(self)

    def __str__(self):
        if self.iana_name is None:
            result = six.next(
                cipher_suite
                for cipher_suite in TlsCipherSuite
                if cipher_suite.value.code == self.code
            ).name
        else:
            result = self.iana_name

        if self.openssl_name:
            result += ' (' + self.openssl_name + ')'

        return result


@attr.s(frozen=True)
class CipherSuiteParams(CipherParamsBase):
    @classmethod
    def get_code_size(cls):
        return 2


TlsCipherSuite = CryptoDataEnumCodedBase('TlsCipherSuite', CryptoDataEnumCodedBase.get_json_records(CipherSuiteParams))


@attr.s(frozen=True)
class CipherKindParams(CipherParamsBase):
    @classmethod
    def get_code_size(cls):
        return 3


SslCipherKind = CryptoDataEnumCodedBase('SslCipherKind', CryptoDataEnumCodedBase.get_json_records(CipherKindParams))


class CipherSuiteExtensionParams(CryptoDataParamsEnumNumeric):
    @classmethod
    def get_code_size(cls):
        return 2


TlsCipherSuiteExtension = CryptoDataEnumCodedBase(
    'TlsCipherSuiteExtension', CryptoDataEnumCodedBase.get_json_records(CipherSuiteExtensionParams)
)


class GreaseOneByteParams(CryptoDataParamsEnumNumeric):
    @classmethod
    def get_code_size(cls):
        return 1


TlsGreaseOneByte = CryptoDataEnumCodedBase(
    'TlsGreaseOneByte', CryptoDataEnumCodedBase.get_json_records(GreaseOneByteParams)
)


class GreaseTwoByteParams(CryptoDataParamsEnumNumeric):
    @classmethod
    def get_code_size(cls):
        return 2


TlsGreaseTwoByte = CryptoDataEnumCodedBase(
    'TlsGreaseTwoByte', CryptoDataEnumCodedBase.get_json_records(GreaseTwoByteParams)
)



# -*- coding: utf-8 -*-

import collections
import datetime
import enum
import re
import six

import attr

from cryptodatahub.common.algorithm import Hash
from cryptodatahub.common.entity import Entity, EntityRole
from cryptodatahub.common.exception import InvalidValue
from cryptodatahub.common.key import PublicKeyX509Base
from cryptodatahub.common.types import (
    Base64Data,
    CryptoDataEnumBase,
    CryptoDataParamsBase,
    CryptoDataParamsFetchedBase,
    convert_base64_data,
    convert_datetime,
    convert_dict_to_object,
    convert_enum,
    convert_iterable,
)
from cryptodatahub.common.utils import bytes_to_hex_string, name_to_enum_item_name


@attr.s(frozen=True)
class CertificateTransparencyOperator(CryptoDataParamsBase):
    name = attr.ib(validator=attr.validators.instance_of(six.string_types))
    email = attr.ib(
        default=(),
        validator=attr.validators.deep_iterable(attr.validators.instance_of(six.string_types))
    )


CertificateTransparencyLogType = enum.Enum('CertificateTransparencyLogType', 'TEST PROD')
CertificateTransparencyLogStateType = enum.Enum(
    'CertificateTransparencyLogStateType', 'PENDING QUALIFIED USABLE READONLY RETIRED REJECTED'
)


class CertificateTransparencyLogDateTimeBase(CryptoDataParamsBase):
    def _asdict_serializer(self, _, __, value):
        if isinstance(value, datetime.datetime):
            return value.strftime("%Y-%m-%dT%H:%M:%SZ")

        return super(CertificateTransparencyLogDateTimeBase, self)._asdict_serializer(_, __, value)


@attr.s(frozen=True)
class CertificateTransparencyLogState(CertificateTransparencyLogDateTimeBase):
    state_type = attr.ib(
        converter=convert_enum(CertificateTransparencyLogStateType),
        validator=attr.validators.optional(attr.validators.instance_of(CertificateTransparencyLogStateType))
    )
    timestamp = attr.ib(
        converter=convert_datetime(),
        validator=attr.validators.instance_of(datetime.datetime),
    )


@attr.s(frozen=True)
class CertificateTransparencyLogTemporalInterval(CertificateTransparencyLogDateTimeBase):
    start_inclusive = attr.ib(
        converter=convert_datetime(),
        validator=attr.validators.instance_of(datetime.datetime),
    )
    end_exclusive = attr.ib(
        converter=convert_datetime(),
        validator=attr.validators.instance_of(datetime.datetime),
    )


@attr.s(frozen=True)
class CertificateTransparencyLogParamsBase(CryptoDataParamsBase):
    log_id = attr.ib(
        converter=convert_base64_data(),
        validator=attr.validators.instance_of(Base64Data),
        metadata={'human_readable_name': 'ID'},
    )


class CertificateTransparencyLogUnknown(CertificateTransparencyLogParamsBase):
    def __str__(self):
        return str(self.log_id)


@attr.s(frozen=True)
class CertificateTransparencyLogParams(  # pylint: disable=too-many-instance-attributes
        CertificateTransparencyLogParamsBase
):
    operator = attr.ib(
        converter=convert_enum(Entity),
        validator=attr.validators.instance_of(Entity),
        metadata={'human_friendly': False},
    )
    key = attr.ib(
        converter=convert_base64_data(),
        validator=attr.validators.instance_of(Base64Data),
        metadata={'human_friendly': False},
    )
    url = attr.ib(
        validator=attr.validators.instance_of(six.string_types),
        metadata={'human_readable_name': 'URL', 'human_friendly': False},
    )
    mmd = attr.ib(
        validator=attr.validators.instance_of(int),
        metadata={'human_readable_name': 'Maximum Merge Delay', 'human_friendly': False},
    )
    description = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(six.string_types)),
    )
    dns = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(six.string_types)),
        metadata={'human_friendly': False},
    )
    temporal_interval = attr.ib(
        default=None,
        converter=convert_dict_to_object(CertificateTransparencyLogTemporalInterval),
        validator=attr.validators.optional(attr.validators.instance_of(CertificateTransparencyLogTemporalInterval)),
        metadata={'human_readable_name': 'DNS', 'human_friendly': False},
    )
    log_type = attr.ib(
        default=None,
        converter=convert_enum(CertificateTransparencyLogType),
        validator=attr.validators.optional(attr.validators.instance_of(CertificateTransparencyLogType)),
        metadata={'human_friendly': False},
    )
    log_state = attr.ib(
        default=None,
        converter=convert_dict_to_object(CertificateTransparencyLogState),
        validator=attr.validators.optional(attr.validators.instance_of(CertificateTransparencyLogState))
    )

    def __attrs_post_init__(self):
        if self.mmd < 1:
            raise ValueError(self.mmd)

    def __str__(self):
        return '{} ({})'.format(
            self.description, self.log_id
        )

    @classmethod
    def description_to_enum_item_name(cls, description):
        name = name_to_enum_item_name(description)

        name = re.sub('([^_])(20[12][0-9][_0-9]*)(H[1-2])?(_LOG)?$', '\\1_\\2\\3\\4', name)

        return name

    @property
    def identifier(self):
        return self.description_to_enum_item_name(self.description)


class CertificateTransparencyLogBase(CryptoDataEnumBase):
    @classmethod
    def from_log_id(cls, log_id):
        log_id = convert_base64_data()(log_id)
        try:
            log = cls._from_attr('log_id', log_id).value
        except InvalidValue:
            log = CertificateTransparencyLogUnknown(log_id)

        return log


CertificateTransparencyLog = CertificateTransparencyLogBase(
    'CertificateTransparencyLog', CryptoDataEnumBase.get_json_records(CertificateTransparencyLogParams)
)


RootCertificateTrustConstraintAction = enum.Enum('RootCertificateTrustConstraintAction', 'DISTRUST DISTURB')


@attr.s(frozen=True)
class CertificateTrustConstraint(object):
    action = attr.ib(
        converter=convert_enum(RootCertificateTrustConstraintAction),
        validator=attr.validators.instance_of(RootCertificateTrustConstraintAction),
    )
    domains = attr.ib(
        default=(),
        validator=attr.validators.deep_iterable(attr.validators.instance_of(six.string_types))
    )
    date = attr.ib(
        default=None,
        converter=convert_datetime('%Y-%m-%dT%H:%M:%S'),
        validator=attr.validators.optional(attr.validators.instance_of(datetime.datetime))
    )


@attr.s(repr=False, slots=True, hash=True)
class _RootCertificateParamCertificateConverter(object):
    def __call__(self, value):
        if value is None:
            return None

        if isinstance(value, PublicKeyX509Base):
            return value

        if not isinstance(value, (list, tuple)):
            return value

        try:
            return PublicKeyX509Base.from_pem_lines(value)
        except ValueError:
            pass

        return value

    def __repr__(self):
        return '<root certificate parameter converter>'


def convert_root_certificate_params():
    return _RootCertificateParamCertificateConverter()


@attr.s(frozen=True)
class RootCertificateTrustStoreConstraint(CryptoDataParamsBase):
    owner = attr.ib(
        converter=convert_enum(Entity),
        validator=attr.validators.instance_of(Entity)
    )
    constraints = attr.ib(
        default=(),
        converter=convert_iterable(convert_dict_to_object(CertificateTrustConstraint)),
        validator=attr.validators.deep_iterable(attr.validators.instance_of(CertificateTrustConstraint))
    )


@attr.s(frozen=True)
class RootCertificateParams(CryptoDataParamsFetchedBase):
    certificate = attr.ib(
        converter=convert_root_certificate_params(),
        validator=attr.validators.instance_of(PublicKeyX509Base)
    )
    trust_stores = attr.ib(
        default=(),
        converter=convert_iterable(convert_dict_to_object(RootCertificateTrustStoreConstraint)),
        validator=attr.validators.deep_iterable(attr.validators.instance_of(RootCertificateTrustStoreConstraint))
    )

    @classmethod
    def subject_to_enum_item_name(cls, subject, serial_number):
        if 'common_name' in subject:
            name = subject['common_name']
        elif 'organizational_unit_name' in subject:
            name = subject['organizational_unit_name']
        else:
            name = subject['organization_name']

        if not isinstance(name, six.string_types):
            name = ' '.join(name)

        return '{}_{}'.format(name_to_enum_item_name(name), serial_number)

    @property
    def identifier(self):
        return self.subject_to_enum_item_name(self.certificate.subject, self.certificate.serial_number)

    def _asdict(self):
        dict_value = super(RootCertificateParams, self)._asdict()
        dict_value['certificate'] = self.certificate.pem.splitlines()
        meta = collections.OrderedDict([
            ('_subject', self.certificate.subject),
            ('_fingerprints', collections.OrderedDict([
                (hash_algorithm.name, fingerprint)
                for hash_algorithm, fingerprint in self.certificate.fingerprints.items()
            ])),
        ])

        return collections.OrderedDict([('_meta', meta)] + list(dict_value.items()))

    def get_constraints_by_owner(self, owner):
        for trust_store_constraint in self.trust_stores:
            if trust_store_constraint.owner == owner:
                return trust_store_constraint.constraints

        raise KeyError(owner)


class RootCertificateBase(CryptoDataEnumBase):
    @classmethod
    def get_json_encoding(cls):
        return 'utf-8'

    @classmethod
    def get_item_by_sha2_256_fingerprint(cls, fingerprint_value):
        if not hasattr(cls, '_ITEMS_BY_SHA2_256_HASH'):
            cls._ITEMS_BY_SHA2_256_HASH = {
                bytes_to_hex_string(item.value.certificate.get_digest(Hash.SHA2_256)): item
                for item in cls
            }

        return cls._ITEMS_BY_SHA2_256_HASH[fingerprint_value.upper().replace(':', '')]

    @classmethod
    def get_items_by_trust_owner(cls, trust_owner):
        if not hasattr(cls, '_ITEMS_BY_TRUST_OWNER'):
            cls._ITEMS_BY_TRUST_OWNER = {
                _trust_owner: []
                for _trust_owner in Entity.get_items_by_role(EntityRole.CA_TRUST_STORE_OWNER)
            }

            for root_certificate in cls:
                for _trust_owner in map(lambda trust_store: trust_store.owner, root_certificate.value.trust_stores):
                    cls._ITEMS_BY_TRUST_OWNER[_trust_owner].append(root_certificate)

            cls._ITEMS_BY_TRUST_OWNER = cls._ITEMS_BY_TRUST_OWNER

        return cls._ITEMS_BY_TRUST_OWNER[trust_owner]


RootCertificate = RootCertificateBase(
    'RootCertificate', RootCertificateBase.get_json_records(RootCertificateParams)
)



# -*- coding: utf-8 -*-

import attr

from cryptodatahub.common.types import CryptoDataEnumBase, CryptoDataParamsEnumNumeric


@attr.s
class VersionParams(CryptoDataParamsEnumNumeric):
    code = attr.ib(validator=attr.validators.instance_of(int))

    @classmethod
    def get_code_size(cls):
        return 2


TlsVersion = CryptoDataEnumBase('TlsVersion', CryptoDataEnumBase.get_json_records(VersionParams))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import codecs
import os
import unittest

from setuptools import setup

from cryptodatahub import __setup__


this_directory = os.getenv('REQUIREMENTS_DIR', '')
with open(os.path.join(this_directory, 'requirements.txt')) as f:
    install_requirements = f.read().splitlines()
this_directory = os.path.abspath(os.path.dirname(__file__))
with codecs.open(os.path.join(this_directory, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()


def test_discover():
    test_loader = unittest.TestLoader()
    test_suite = test_loader.discover('test', pattern='test_*.py')
    return test_suite


setup(
    name=__setup__.__title__,
    version=__setup__.__version__,
    description=__setup__.__description__,
    long_description=long_description,
    long_description_content_type='text/x-rst',
    author=__setup__.__author__,
    author_email=__setup__.__author_email__,
    maintainer=__setup__.__author__,
    maintainer_email=__setup__.__author_email__,
    license=__setup__.__license__,
    license_files=['LICENSE.txt', ],
    project_urls={
        'Homepage': __setup__.__url__,
        'Changelog': 'https://' + __setup__.__technical_name__ + '.readthedocs.io/en/latest/changelog',
        'Documentation': 'https://' + __setup__.__technical_name__ + '.readthedocs.io/en/latest/',
        'Issues': __setup__.__url__ + '/-/issues',
        'Source': __setup__.__url__,
    },
    keywords='ssl tls ssh dns',

    install_requires=install_requirements,
    extras_require={
        ":python_version < '3'": [
            "enum34==1.1.6",
            "pathlib2==2.3.7.post1",
            "Mock",
            "unittest2",
        ],
        "test": ["coverage", ],
        "pep8": ["flake8", ],
        "pylint": ["pylint", ],
    },

    packages=[
        'cryptodatahub',
        'cryptodatahub.common',
        'cryptodatahub.dnsrec',
        'cryptodatahub.ssh',
        'cryptodatahub.tls',
    ],

    package_data={
        'cryptodatahub.common': ['*.json'],
        'cryptodatahub.dnsrec': ['*.json'],
        'cryptodatahub.ssh': ['*.json'],
        'cryptodatahub.tls': ['*.json'],
    },

    test_suite='setup.test_discover',

    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Framework :: tox',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Science/Research',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)',
        'Natural Language :: English',
        'Operating System :: MacOS',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Programming Language :: Python',
        'Topic :: Internet',
        'Topic :: Security',
        'Topic :: Security :: Cryptography',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)



# -*- coding: utf-8 -*-

import abc
import base64
import collections
import datetime
import os

from collections import OrderedDict

import six
import asn1crypto.keys
import asn1crypto.pem
import asn1crypto.x509
import attr

from cryptodatahub.common.algorithm import Authentication, KeyExchange, Hash, NamedGroup, Signature
from cryptodatahub.common.exception import InvalidValue
from cryptodatahub.common.grade import (
    AttackNamed,
    AttackType,
    Grade,
    GradeableComplex,
    GradeableVulnerabilities,
    Vulnerability,
)
from cryptodatahub.common.types import _ConverterBase
from cryptodatahub.common.utils import bytes_to_hex_string, hash_bytes

from cryptodatahub.tls.algorithm import TlsExtensionType


@attr.s(frozen=True)
class _PublicKeySizeGradeable(GradeableVulnerabilities):
    @classmethod
    def get_gradeable_name(cls):
        return 'public key size'


@attr.s(frozen=True)
class PublicKeySize(GradeableComplex):
    _FINITE_FIELD_TYPES = [Authentication.RSA, Authentication.DSS, KeyExchange.ADH, KeyExchange.DH, KeyExchange.DHE]
    _ELLIPTIC_CURVE_TYPES = [Authentication.ECDSA, Authentication.EDDSA, KeyExchange.ECDH, KeyExchange.ECDHE]

    key_type = attr.ib(validator=attr.validators.instance_of((Authentication, KeyExchange)))
    value = attr.ib(validator=attr.validators.instance_of(six.integer_types))

    @value.validator
    def _value_validator(self, attribute, value):  # pylint: disable=unused-argument
        if value <= 0:
            raise InvalidValue(value, type(self), 'value')

    def __attrs_post_init__(self):
        if self.key_type in self._ELLIPTIC_CURVE_TYPES:
            gradeables = []
            attack_type = AttackType.DISCRETE_LOGARITHM
            if self.value <= 112:
                gradeables.append(Vulnerability(attack_type=attack_type, grade=Grade.INSECURE, named=None))
            elif self.value <= 160:
                gradeables.append(Vulnerability(attack_type=attack_type, grade=Grade.WEAK, named=None))
            gradeables = [_PublicKeySizeGradeable(gradeables)]
        elif self.key_type in self._FINITE_FIELD_TYPES:
            gradeables = []
            attack_type = AttackType.INTEGER_FACTORIZATION
            if self.value <= 768:
                if self.key_type == Authentication.RSA:
                    attack_named = AttackNamed.FREAK
                elif isinstance(self.key_type, KeyExchange):
                    attack_named = AttackNamed.WEAK_DH
                else:
                    attack_named = None

                gradeables.append(Vulnerability(attack_type=attack_type, grade=Grade.INSECURE, named=attack_named))
            elif self.value <= 1024:
                gradeables.append(Vulnerability(attack_type=attack_type, grade=Grade.WEAK, named=None))
            elif self.key_type in [KeyExchange.ADH, KeyExchange.DH, KeyExchange.DHE] and self.value > 4096:
                gradeables.append(
                    Vulnerability(attack_type=AttackType.DOS_ATTACK, grade=Grade.WEAK, named=AttackNamed.DHEAT_ATTACK)
                )
            gradeables = [_PublicKeySizeGradeable(gradeables)]
        elif self.key_type == KeyExchange.HYBRID_PQS:
            gradeables = []
        else:
            gradeables = None

        object.__setattr__(self, 'gradeables', gradeables)

    def __str__(self):
        return str(self.value)


@attr.s(repr=False, slots=True, hash=True)
class _PublicKeySizeConverter(_ConverterBase):
    key_exchange = attr.ib(validator=attr.validators.instance_of(KeyExchange))

    def __call__(self, value):
        if value is None:
            return None

        try:
            return PublicKeySize(self.key_exchange, value)
        except (TypeError, InvalidValue):
            return value

    def __repr__(self):
        return '<public key size converter>'


def convert_public_key_size(key_exchange):
    return _PublicKeySizeConverter(key_exchange)


@attr.s(frozen=True)
class PublicKeyParamBase(object):
    pass


@attr.s(frozen=True)
class PublicKeyParamsDsa(PublicKeyParamBase):
    prime = attr.ib(validator=attr.validators.instance_of(six.integer_types))
    generator = attr.ib(validator=attr.validators.instance_of(six.integer_types))
    order = attr.ib(validator=attr.validators.instance_of(six.integer_types))
    public_key_value = attr.ib(validator=attr.validators.instance_of(six.integer_types))


@attr.s(frozen=True)
class PublicKeyParamsEcdsa(PublicKeyParamBase):
    named_group = attr.ib(validator=attr.validators.instance_of(NamedGroup))
    point_x = attr.ib(validator=attr.validators.instance_of(six.integer_types))
    point_y = attr.ib(validator=attr.validators.instance_of(six.integer_types))

    @classmethod
    def from_octet_bit_string(cls, named_group, octet_bit_string):
        point_x, point_y = asn1crypto.keys.ECPointBitString(bytes(octet_bit_string)).to_coords()
        return cls(
            named_group,
            point_x,
            point_y,
        )

    @property
    def octet_bit_string(self):
        return bytes(asn1crypto.keys.ECPointBitString.from_coords(self.point_x, self.point_y))


@attr.s(frozen=True)
class PublicKeyParamsEddsa(PublicKeyParamBase):
    curve_type = attr.ib(validator=attr.validators.instance_of(NamedGroup))
    key_data = attr.ib(validator=attr.validators.instance_of((bytes, bytearray)))


@attr.s(frozen=True)
class PublicKeyParamsRsa(PublicKeyParamBase):
    modulus = attr.ib(validator=attr.validators.instance_of(six.integer_types))
    public_exponent = attr.ib(validator=attr.validators.instance_of(six.integer_types))


@attr.s(eq=False, frozen=True)
class PublicKey(object):
    _public_key = attr.ib(validator=attr.validators.instance_of(asn1crypto.keys.PublicKeyInfo))

    @classmethod
    def from_der(cls, der):
        return cls(asn1crypto.keys.PublicKeyInfo.load(bytes(der)))

    @classmethod
    def from_pem(cls, pem):
        return cls.from_der(asn1crypto.pem.unarmor(pem.encode('ascii'))[2])

    @classmethod
    def from_pem_lines(cls, pem_lines):
        return cls.from_pem(os.linesep.join(pem_lines))

    @classmethod
    def from_params(cls, params):
        if isinstance(params, PublicKeyParamsDsa):
            algorithm_id = asn1crypto.keys.PublicKeyAlgorithmId(six.u('dsa'))
            parameters = asn1crypto.keys.DSAParams({
                'p': params.prime, 'g': params.generator, 'q': params.order,
            })
            public_key = asn1crypto.keys.PublicKeyInfo({
                'algorithm': asn1crypto.keys.PublicKeyAlgorithm({
                    'algorithm': algorithm_id,
                    'parameters': parameters,
                }),
                'public_key': params.public_key_value,
            })
        elif isinstance(params, PublicKeyParamsEddsa):
            if params.curve_type == NamedGroup.CURVE25519:
                algorithm_name = six.u('ed25519')
            elif params.curve_type == NamedGroup.CURVE448:
                algorithm_name = six.u('ed448')
            else:
                raise NotImplementedError()

            algorithm_id = asn1crypto.keys.PublicKeyAlgorithmId(algorithm_name)
            public_key = asn1crypto.keys.PublicKeyInfo({
                'algorithm': asn1crypto.keys.PublicKeyAlgorithm({
                    'algorithm': algorithm_id,
                }),
                'public_key': asn1crypto.core.OctetBitString(bytes(params.key_data)),
            })
        elif isinstance(params, PublicKeyParamsRsa):
            algorithm_id = asn1crypto.keys.PublicKeyAlgorithmId(six.u('rsa'))
            public_key = asn1crypto.keys.PublicKeyInfo({
                'algorithm': asn1crypto.keys.PublicKeyAlgorithm({'algorithm': algorithm_id}),
                'public_key': asn1crypto.keys.RSAPublicKey(
                    attr.asdict(params, recurse=False, value_serializer=None)
                )
            })
        elif isinstance(params, PublicKeyParamsEcdsa):
            algorithm_id = asn1crypto.keys.PublicKeyAlgorithmId(six.u('ec'))
            parameters = asn1crypto.keys.ECDomainParameters({
                'named': params.named_group.value.oid
            })

            public_key = asn1crypto.keys.PublicKeyInfo({
                'algorithm': asn1crypto.keys.PublicKeyAlgorithm({
                    'algorithm': algorithm_id,
                    'parameters': parameters,
                }),
                'public_key': asn1crypto.keys.ECPointBitString.from_coords(params.point_x, params.point_y)
            })
        else:
            raise NotImplementedError(type(params))

        return cls(public_key)

    @property
    def params(self):
        if self.key_type == Authentication.DSS:
            public_key = self._public_key['public_key'].parsed
            parameters = self._public_key['algorithm']['parameters']
            return PublicKeyParamsDsa(
                prime=parameters['p'].native,
                generator=parameters['g'].native,
                order=parameters['q'].native,
                public_key_value=public_key.native,
            )
        if self.key_type == Authentication.ECDSA:
            public_key = self._public_key['public_key']
            parameters = self._public_key['algorithm']['parameters']

            return PublicKeyParamsEcdsa(
                NamedGroup.from_oid(parameters.chosen.dotted),
                *public_key.to_coords()
            )
        if self.key_type == Authentication.EDDSA:
            algorithm = self._public_key['algorithm']['algorithm']
            signature = Signature.from_oid(algorithm.dotted)
            if signature == Signature.ED25519:
                curve_type = NamedGroup.CURVE25519
            elif signature == Signature.ED448:
                curve_type = NamedGroup.CURVE448
            else:
                raise NotImplementedError()

            return PublicKeyParamsEddsa(
                curve_type=curve_type,
                key_data=self._public_key['public_key'].native
            )
        if self.key_type == Authentication.RSA:
            public_key = self._public_key['public_key'].parsed
            return PublicKeyParamsRsa(
                modulus=public_key['modulus'].native,
                public_exponent=public_key['public_exponent'].native,
            )

        raise NotImplementedError(self.key_type)

    @classmethod
    def _get_type_name(cls):
        return 'public key'

    def __eq__(self, other):
        return self.der == other.der

    @property
    def der(self):
        return self._public_key.dump()

    @property
    def pem(self):
        return six.ensure_str(asn1crypto.pem.armor(six.u(self._get_type_name().upper()), self.der))

    @property
    def key_type(self):
        try:
            key_type_oid = self._public_key['algorithm']['algorithm'].dotted
        except KeyError as e:
            key_type_oid = e.args[0]

        try:
            return Signature.from_oid(key_type_oid).value.key_type
        except InvalidValue:
            return Authentication.from_oid(key_type_oid)

    @property
    def key_size(self):
        if self.key_type == Authentication.GOST_R3410_12_256:
            return 256
        if self.key_type == Authentication.GOST_R3410_12_512:
            return 512
        if self.key_type == Authentication.GOST_R3410_01:
            return 256
        if self.key_type == Authentication.EDDSA:
            return self.params.curve_type.value.size

        return int(self._public_key.bit_size)

    @property
    def key_bytes(self):
        return PublicKey.der.fget(self)

    def get_digest(self, hash_type):
        return hash_bytes(hash_type, self.der)

    def fingerprint(self, hash_type):
        return bytes_to_hex_string(self.get_digest(hash_type), ':')

    @property
    def fingerprints(self):
        return OrderedDict([
            (hash_type, self.fingerprint(hash_type))
            for hash_type in [Hash.MD5, Hash.SHA1, Hash.SHA2_256]
        ])

    def _asdict(self):
        return collections.OrderedDict([
            ('algorithm', self.key_type),
            ('size', PublicKeySize(self.key_type, self.key_size)),
            ('fingerprints', self.fingerprints),
        ])


class PublicKeySigned(PublicKey):
    @property
    @abc.abstractmethod
    def valid_not_before(self):
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def valid_not_after(self):
        raise NotImplementedError()

    @property
    def validity_period(self):
        return self.valid_not_after - self.valid_not_before

    @property
    def validity_remaining_time(self):
        now = datetime.datetime.now(asn1crypto.util.timezone.utc)
        return self.valid_not_after - now if now < self.valid_not_after else None

    @property
    def expired(self):
        return datetime.datetime.now(asn1crypto.util.timezone.utc) > self.valid_not_after

    @property
    @abc.abstractmethod
    def signature_hash_algorithm(self):
        raise NotImplementedError()


@attr.s(eq=False, init=False, frozen=True)
class PublicKeyX509Base(PublicKeySigned):  # pylint: disable=too-many-public-methods
    _EV_OIDS_BY_CA = {
        'A-Trust': ('1.2.40.0.17.1.22', ),
        'Actalis': ('1.3.159.1.17.1', ),
        'AffirmTrust': (
            '1.3.6.1.4.1.34697.2.1',
            '1.3.6.1.4.1.34697.2.2',
            '1.3.6.1.4.1.34697.2.3',
            '1.3.6.1.4.1.34697.2.4',
        ),
        'Buypass': ('2.16.578.1.26.1.3.3', ),
        'Camerfirma': (
            '1.3.6.1.4.1.17326.10.14.2.1.2',
            '1.3.6.1.4.1.17326.10.8.12.1.2',
        ),
        'Comodo Group': ('1.3.6.1.4.1.6449.1.2.1.5.1', ),
        'DigiCert': (
            '2.16.840.1.114412.1.3.0.2',
            '2.16.840.1.114412.2.1',
        ),
        'DigiNotar': ('2.16.528.1.1001.1.1.1.12.6.1.1.1', ),
        'E-Tugra': ('2.16.792.3.0.4.1.1.4', ),
        'ETSI': (
            '0.4.0.2042.1.4',
            '0.4.0.2042.1.5',
        ),
        'Entrust': ('2.16.840.1.114028.10.1.2', ),
        'Firmaprofesional': ('1.3.6.1.4.1.13177.10.1.3.10', ),
        'GeoTrust': ('1.3.6.1.4.1.14370.1.6', ),
        'GlobalSign': ('1.3.6.1.4.1.4146.1.1', ),
        'Go Daddy': ('2.16.840.1.114413.1.7.23.3', ),
        'Izenpe': ('1.3.6.1.4.1.14777.6.1.1', ),
        'Kamu Sertifikasyon Merkezi': ('2.16.792.1.2.1.1.5.7.1.9', ),
        'Logius PKIoverheid': ('2.16.528.1.1003.1.2.7', ),
        'Network Solutions': ('1.3.6.1.4.1.782.1.2.1.8.1', ),
        'OpenTrust/DocuSign France': ('1.3.6.1.4.1.22234.2.5.2.3.1', ),
        'QuoVadis': ('1.3.6.1.4.1.8024.0.2.100.1.2', ),
        'SECOM Trust Systems': ('1.2.392.200091.100.721.1', ),
        'SHECA': ('1.2.156.112570.1.1.3', ),
        'Starfield Technologies': ('2.16.840.1.114414.1.7.23.3', ),
        'StartCom Certification Authority': (
            '1.3.6.1.4.1.23223.1.1.1',
            '1.3.6.1.4.1.23223.2',
        ),
        'SwissSign': ('2.16.756.1.89.1.2.1.1', ),
        'Swisscom': ('2.16.756.1.83.21.0', ),
        'Symantec (VeriSign)': ('2.16.840.1.113733.1.7.23.6', ),
        'T-Systems': ('1.3.6.1.4.1.7879.13.24.1', ),
        'Thawte': ('2.16.840.1.113733.1.7.48.1', ),
        'Trustwave': ('2.16.840.1.114404.1.1.2.4.1', ),
        'Verizon Business (formerly Cybertrust)': ('1.3.6.1.4.1.6334.1.100.1', ),
        'Wells Fargo': ('2.16.840.1.114171.500.9', ),
        'WoSign': ('1.3.6.1.4.1.36305.2', ),
    }

    _certificate = attr.ib(validator=attr.validators.instance_of(asn1crypto.x509.Certificate))

    def __init__(self, certificate):
        super(PublicKeySigned, self).__init__(certificate.public_key)

        object.__setattr__(self, '_certificate', certificate)

    @classmethod
    def _get_type_name(cls):
        return 'certificate'

    @classmethod
    def from_der(cls, der):
        return cls(asn1crypto.x509.Certificate.load(bytes(der)))

    @property
    def der(self):
        return self._certificate.dump()

    @property
    def valid_not_before(self):
        return self._certificate.not_valid_before

    @property
    def valid_not_after(self):
        return self._certificate.not_valid_after

    @property
    def signature_hash_algorithm(self):
        try:
            signature_oid = self._certificate['signature_algorithm']['algorithm'].dotted
        except KeyError as e:
            signature_oid = e.args[0]

        return Signature.from_oid(signature_oid)

    @property
    def public_key(self):
        return PublicKey(self._public_key)

    @property
    def public_key_pin(self):
        return base64.b64encode(hash_bytes(Hash.SHA2_256, self.key_bytes)).decode('ascii')

    def _has_any_policy_value(self, oid_values):
        if self._certificate.certificate_policies_value is None:
            return False

        for policy_information in self._certificate.certificate_policies_value:
            if policy_information['policy_identifier'].dotted in oid_values:
                return True

        return False

    @property
    def extended_validation(self):
        return any(map(self._has_any_policy_value, self._EV_OIDS_BY_CA.values()))

    @property
    def tls_features(self):
        tls_feature_value = self._certificate.tls_feature_value
        if tls_feature_value is None:
            return []

        return list(map(lambda feature: TlsExtensionType.from_code(feature.native), tls_feature_value))

    @property
    def serial_number(self):
        return self._certificate.serial_number

    @property
    def subject(self):
        return OrderedDict([
            (six.ensure_str(name), value)
            for name, value in self._certificate.subject.native.items()
        ])

    @property
    def issuer(self):
        return self._certificate.issuer.native

    @property
    def valid_domains(self):
        return self._certificate.valid_domains

    def is_subject_matches(self, host_name):
        return self._certificate.is_valid_domain_ip(six.u(host_name))

    @property
    def subject_alternative_names(self):
        if self._certificate.subject_alt_name_value is None:
            return []

        return self._certificate.subject_alt_name_value.native

    @property
    def crl_distribution_points(self):
        if self._certificate.crl_distribution_points_value is None:
            return []

        return [
            crl_distribution_point.url
            for crl_distribution_point in self._certificate.crl_distribution_points_value
        ]

    @property
    def ocsp_responders(self):
        return self._certificate.ocsp_urls

    @property
    def is_ca(self):
        return self._certificate.ca

    @property
    def is_self_signed(self):
        return self._certificate.self_issued

    def _asdict(self):
        items = [
            ('algorithm', self.key_type),
            ('size', PublicKeySize(self.key_type, self.key_size)),
            ('version', self._certificate['tbs_certificate']['version'].native),
            ('serial_number', str(self.serial_number)),
            ('subject', self.subject),
            ('subject_alternative_names', sorted(self.subject_alternative_names)),
            ('issuer', self.issuer),
            ('signature_hash_algorithm', self.signature_hash_algorithm),
            ('validity', collections.OrderedDict([
                ('not_before', str(self.valid_not_before)),
                ('not_after', str(self.valid_not_after)),
                ('period', str(self.validity_period)),
                ('remaining', str(self.validity_remaining_time.days) if self.validity_remaining_time else None),
            ])),
            ('revocation', collections.OrderedDict([
                ('crl_distribution_points', self.crl_distribution_points),
                ('ocsp_responders', self.ocsp_responders),
            ])),
            ('fingerprints', self.fingerprints),
            ('public_key_pin', self.public_key_pin),
        ]

        if not self.is_ca:
            items += [
                ('end_entity', collections.OrderedDict([
                    ('extended_validation', self.extended_validation),
                    ('tls_features', list(map(lambda feature: feature.name, self.tls_features))),
                ]))
            ]

        return collections.OrderedDict(items)



# -*- coding: utf-8 -*-

import six
import attr

from cryptodatahub.common.entity import Entity, Server
from cryptodatahub.common.exception import InvalidValue
from cryptodatahub.common.grade import GradeableVulnerabilities
from cryptodatahub.common.types import (
    CryptoDataEnumBase,
    CryptoDataParamsBase,
    CryptoDataParamsNamed,
    convert_big_enum,
    convert_dict_to_object,
    convert_enum,
    convert_iterable,
    convert_variadic,
)


@attr.s(frozen=True)
class StandardParams(CryptoDataParamsNamed):
    publisher = attr.ib(
        converter=convert_enum(Entity),
        validator=attr.validators.instance_of(Entity),
    )


Standard = CryptoDataEnumBase('Standard', CryptoDataEnumBase.get_json_records(StandardParams))


@attr.s(eq=False, frozen=True)
class DHParameterNumbers(object):
    p = attr.ib(  # pylint: disable=invalid-name
        converter=convert_big_enum(),
        validator=attr.validators.instance_of(six.integer_types),
        metadata={'human_readable_name': 'p'},
    )
    g = attr.ib(  # pylint: disable=invalid-name
        converter=convert_big_enum(),
        validator=attr.validators.instance_of(six.integer_types),
        metadata={'human_readable_name': 'g'},
    )
    q = attr.ib(  # pylint: disable=invalid-name
        default=None,
        converter=convert_big_enum(),
        validator=attr.validators.optional(attr.validators.instance_of(six.integer_types)),
        metadata={'human_readable_name': 'q'},
    )

    def __eq__(self, other):
        return self.p == other.p and self.g == other.g and (self.q is None or self.q == other.q)


@attr.s(eq=False, frozen=True)
class DHParamWellKnownParams(CryptoDataParamsBase, GradeableVulnerabilities):
    parameter_numbers = attr.ib(
        converter=convert_dict_to_object(DHParameterNumbers),
        validator=attr.validators.instance_of(DHParameterNumbers)
    )
    name = attr.ib(validator=attr.validators.instance_of(six.string_types))
    source = attr.ib(
        converter=convert_variadic((convert_enum(Entity), convert_enum(Server))),
        validator=attr.validators.instance_of((Entity, Server))
    )
    standards = attr.ib(
        converter=convert_iterable(convert_enum(Standard)),
        validator=attr.validators.deep_iterable(attr.validators.instance_of(Standard))
    )
    key_size = attr.ib(validator=attr.validators.instance_of(six.integer_types))
    safe_prime = attr.ib(default=True, validator=attr.validators.instance_of(bool))

    @classmethod
    def get_gradeable_name(cls):
        return 'DH parameter'

    def __str__(self):
        if self.standards:
            return '{} ({})'.format(self.name, ', '.join(map(lambda standard: standard.value.name, self.standards)))

        return '{}-bit {} {} DH parameter'.format(self.key_size, self.source.value.name, self.name)


class DHParamWellKnownBase(CryptoDataEnumBase):
    @classmethod
    def from_parameter_numbers(cls, parameter_numbers):
        return cls._from_attr('parameter_numbers', parameter_numbers)


DHParamWellKnown = DHParamWellKnownBase(
    'DHParamWellKnown', DHParamWellKnownBase.get_json_records(DHParamWellKnownParams)
)


@attr.s
class ECParameterNumbers(object):
    a = attr.ib(  # pylint: disable=invalid-name
        converter=convert_big_enum(),
        validator=attr.validators.instance_of(six.integer_types),
        metadata={'human_readable_name': 'a'},
    )
    b = attr.ib(  # pylint: disable=invalid-name
        converter=convert_big_enum(),
        validator=attr.validators.instance_of(six.integer_types),
        metadata={'human_readable_name': 'b'},
    )
    x = attr.ib(  # pylint: disable=invalid-name
        converter=convert_big_enum(),
        validator=attr.validators.instance_of(six.integer_types),
        metadata={'human_readable_name': 'x'},
    )
    y = attr.ib(  # pylint: disable=invalid-name
        converter=convert_big_enum(),
        validator=attr.validators.instance_of(six.integer_types),
        metadata={'human_readable_name': 'y'},
    )


@attr.s(eq=False, frozen=True)
class ECParamWellKnownParams(CryptoDataParamsNamed):
    standards = attr.ib(
        converter=convert_iterable(convert_enum(Standard)),
        validator=attr.validators.deep_iterable(attr.validators.instance_of(Standard))
    )
    aliases = attr.ib(validator=attr.validators.deep_iterable(attr.validators.instance_of(six.string_types)))
    parameter_numbers = attr.ib(
        converter=convert_dict_to_object(ECParameterNumbers),
        validator=attr.validators.instance_of(ECParameterNumbers)
    )


class ECParamWellKnownBase(CryptoDataEnumBase):
    @classmethod
    def from_parameter_numbers(cls, parameter_numbers):
        return cls._from_attr('parameter_numbers', parameter_numbers)

    @classmethod
    def from_named_group(cls, named_group):
        try:
            ec_param = cls[named_group.name]
        except KeyError as e:
            six.raise_from(InvalidValue(named_group, cls), e)

        return ec_param


ECParamWellKnown = ECParamWellKnownBase(
    'ECParamWellKnown', ECParamWellKnownBase.get_json_records(ECParamWellKnownParams)
)



# -*- coding: utf-8 -*-

import binascii
import hashlib

import attr
import six
import urllib3

from cryptodatahub.common.algorithm import Hash


def bytes_to_hex_string(byte_array, separator='', lowercase=False):
    if lowercase:
        format_str = '{:02x}'
    else:
        format_str = '{:02X}'

    return separator.join([format_str.format(x) for x in six.iterbytes(bytes(byte_array))])


def bytes_from_hex_string(hex_string, separator=''):
    if separator:
        hex_string = ''.join(hex_string.split(separator))

    try:
        binary_data = binascii.a2b_hex(hex_string)
    except (TypeError, ValueError) as e:
        six.raise_from(ValueError(*e.args), e)

    return binary_data


def name_to_enum_item_name(name):
    converted_name = ''
    for char in name:
        if char.isalnum():
            converted_name += char
        elif char == '\'':
            pass
        elif converted_name and converted_name[-1] != '_':
            converted_name += '_'

    return converted_name.rstrip('_').upper()


_HASHLIB_FUNCS = {
    Hash.MD5: hashlib.md5,
    Hash.SHA1: hashlib.sha1,
    Hash.SHA2_224: hashlib.sha224,
    Hash.SHA2_256: hashlib.sha256,
    Hash.SHA2_384: hashlib.sha384,
    Hash.SHA2_512: hashlib.sha512,
}


def hash_bytes(hash_algorithm, hashable_value):
    try:
        hashlib_funcs = _HASHLIB_FUNCS[hash_algorithm]
    except KeyError as e:
        six.raise_from(NotImplementedError(hash_algorithm), e)

    return hashlib_funcs(hashable_value).digest()


@attr.s
class HttpFetcher(object):
    connect_timeout = attr.ib(default=2, validator=attr.validators.instance_of((int, float)))
    read_timeout = attr.ib(default=1, validator=attr.validators.instance_of((int, float)))
    retry = attr.ib(default=1, validator=attr.validators.instance_of(int))
    _request_params = attr.ib(default=None, init=False)
    _response = attr.ib(default=None, init=False)

    def __attrs_post_init__(self):
        request_params = {
            'preload_content': False,
            'timeout': urllib3.Timeout(connect=self.connect_timeout, read=self.read_timeout),
            'retries': urllib3.Retry(
                self.retry, status_forcelist=urllib3.Retry.RETRY_AFTER_STATUS_CODES | frozenset([502])
            ),
        }

        object.__setattr__(self, '_request_params', request_params)

    def get_response_header(self, header_name):
        if self._response is None:
            raise AttributeError()

        return self._response.headers.get(header_name, None)

    @property
    def response_data(self):
        if self._response is None:
            raise AttributeError()

        return self._response.data

    def fetch(self, url):
        pool_manager = urllib3.PoolManager()

        try:
            self._response = pool_manager.request('GET', str(url), **self._request_params)
        except BaseException as e:  # pylint: disable=broad-except
            if e.__class__.__name__ != 'TimeoutError' and not isinstance(e, urllib3.exceptions.HTTPError):
                raise e

        pool_manager.clear()

    def __call__(self, url):
        self.fetch(url)

        return self.response_data
